@@ -26,12 +26,12 @@ cmpthese(
     {
         pkg_caix          => sub { AAA->a },
         pkg_cag           => sub { AAA2->a },
-        pkg_gparent_caixs => sub { CCC->a },
+        pkg_gparent_caix  => sub { CCC->a },
         pkg_gparent_cag   => sub { CCC2->a },
         pkg_set_caix      => sub { AAA->a(42) },
-        pkg_set_cag       => sub { AAA2->a(42) },
-#        caixs_pset        => sub { $o->caixc(42) },
-#        caixs_pget        => sub { $o->caixc },
+#        pkg_set_cag       => sub { AAA2->a(42) },
+#        class_set_caix    => sub { $o->caixc(42) },
+        class_caix        => sub { $o->caixc },
         obj_caix          => sub { $o->a },
         obj_caix_cb       => sub { $o->caxsi },
         obj_cag           => sub { $o2->a },
